#!/usr/bin/env python

import argparse

from brother_ql.models import ModelsManager
from brother_ql.labels import FormFactor, LabelsManager

def main():
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers(dest='action')
    subparser.add_parser('list-label-sizes', help='List available label sizes')
    subparser.add_parser('list-models',      help='List available models')
    args = parser.parse_args()

    if not args.action:
        parser.error('Please choose an action')

    elif args.action == 'list-models':
        print('Supported models:')
        for model in ModelsManager().iter_identifiers(): print(" " + model)

    elif args.action == 'list-label-sizes':
        print('Supported label sizes:')
        fmt = " {label_size:9s} {dots_printable:14s} {label_descr:26s}"
        print(fmt.format(label_size="Name", label_descr="Description", dots_printable="Printable px"))
        for label in LabelsManager().iter_elements():
            if label.form_factor == FormFactor.DIE_CUT:
                label_descr = "(%d x %d mm^2)"  % label.tape_size
                dots_printable = "{0:4d} x {1:4d}".format(label.dots_printable)
            elif label.form_factor == FormFactor.ENDLESS:
                label_descr = "(%d mm endless)" % label.tape_size[0]
                dots_printable = "{0:4d}".format(label.dots_printable)
            elif label.form_factor == FormFactor.ROUND_DIE_CUT:
                label_descr = "(%d mm diameter, round)" % label.tape_size[0]
                dots_printable = "{0:4d} x {1:4d}".format(label.dots_printable)
            else:
                raise Exception("Unknown label form factor: {label.form_factor}")

            print(fmt.format(label_size=label.identifier, label_descr=label_descr, dots_printable=dots_printable))

if __name__ == "__main__": main()



"""
Module containing helper functions for printing information in a human-readable format
"""

import logging
from typing import Sequence

from brother_ql.labels import FormFactor, LabelsManager

logger = logging.getLogger(__name__)

def textual_label_description(label_sizes_to_include: Sequence[str]) -> str:
    """
    Returns a textual description of labels with the specified sizes
    """
    output = "Supported label sizes:\n"
    output = ""
    fmt = " {label_size:9s} {dots_printable:14s} {label_descr:26s}\n"
    output += fmt.format(label_size="Name", dots_printable="Printable px", label_descr="Description")
    #output += fmt.format(label_size="", dots_printable="width x height", label_descr="")
    for label_size in label_sizes_to_include:
        label = LabelsManager()[label_size]
        if label.form_factor in (FormFactor.DIE_CUT, FormFactor.ROUND_DIE_CUT):
            dp_fmt = "{0:4d} x {1:4d}"
        elif label.form_factor == FormFactor.ENDLESS:
            dp_fmt = "{0:4d}"
        else:
            dp_fmt = " - unknown - "
        dots_printable = dp_fmt.format(*label.dots_printable)
        label_descr = label.name
        output += fmt.format(label_size=label_size, dots_printable=dots_printable, label_descr=label_descr)
    return output

def log_discovered_devices(available_devices, level=logging.INFO):
    """
    Logs all automatically discovered devices to console.
    """
    for dev in available_devices:
        result = {'model': 'unknown'}
        result.update(dev)
        logger.log(level, "  Found a label printer: {identifier}  (model: {model})".format(**result))

def textual_description_discovered_devices(available_devices) -> str:
    # TODO: figure out what is the type of the argument and document it
    output = ""
    for dev in available_devices:
        output += dev['identifier']
    return output



"""
This module contains all the classes of exceptions that may be raised by brother_ql
"""

class BrotherQLError(Exception):
    """ Base class for exceptions from this package """
    pass

class BrotherQLUnsupportedCmd(BrotherQLError):
    """ Raised when a raster command is not supported with a given printer/label combination """
    pass

class BrotherQLUnknownModel(BrotherQLError):
    """ Unrecognized printer model """
    pass

class BrotherQLRasterError(BrotherQLError):
    """ Raised when invalid data is passed to functions on ``brother_ql.raster.BrotherQLRaster`` """
    pass




import logging

logger = logging.getLogger(__name__)

class ElementsManager(object):
    """
    A class managing a collection of 'elements'.
    Those elements are expected to be objects that
    * can be compared for equality against each other
    * have the attribute .identifier
    """
    elements = [] #: list of elements contained by this manager
    element_name = "element" #: what an element is

    def iter_identifiers(self):
        """ Returns an iterator over the identifiers of all the elements in this manager """
        for element in self.elements:
            yield element.identifier

    def iter_elements(self):
        """ Returns an iterator over all the element objects in this manager """
        for element in self.elements:
            yield element

    def identifiers(self):
        """ Returns a list of the identifiers of all the elements in this manager """
        return list(map(lambda e : e.identifier, self.elements))

    def get(self, identifier: str, default=None):
        """
        Returns the element object with the given identifier, or if it doesn't exist, the default value

        :param identifier: the identifier to look for
        :param default: what to return if the given identifier is not found
        """
        return next(filter(lambda l : l.identifier == identifier, self.elements), default)

    def __getitem__(self, key):
        value = next(filter(lambda l : l.identifier == key, self.elements), None)
        if value is not None:
            return value
        else:
            raise KeyError()



from attr import attrs, attrib
from typing import Tuple

import copy

from brother_ql.helpers import ElementsManager

@attrs
class Model(object):
    """
    This class represents a printer model. All specifics of a certain model
    and the opcodes it supports should be contained in this class.
    """
    #: A string identifier given to each model implemented. Eg. 'QL-500'.
    identifier: str = attrib()
    #: Minimum and maximum number of rows or 'dots' that can be printed.
    #: Together with the dpi this gives the minimum and maximum length
    #: for continuous tape printing.
    min_max_length_dots: Tuple[int, int] = attrib()
    #: The minimum and maximum amount of feeding a label
    min_max_feed: Tuple[int, int] = attrib(default=(35, 1500))
    number_bytes_per_row: int = attrib(default=90)
    #: The required additional offset from the right side
    additional_offset_r: int = attrib(default=0)
    #: Support for the 'mode setting' opcode
    mode_setting: bool = attrib(default=True)
    #: Model has a cutting blade to automatically cut labels
    cutting: bool = attrib(default=True)
    #: Model has support for the 'expanded mode' opcode.
    #: (So far, all models that have cutting support do).
    expanded_mode: bool = attrib(default=True)
    #: Model has support for compressing the transmitted raster data.
    #: Some models with only USB connectivity don't support compression.
    compression: bool = attrib(default=True)
    #: Support for two color printing (black/red/white)
    #: available only on some newer models.
    two_color: bool = attrib(default=False)
    #: Number of NULL bytes needed for the invalidate command.
    num_invalidate_bytes: int = attrib(default=200)

    @property
    def name(self) -> str:
        """
        Returns the printer identifier (already human-readable)
        """
        return self.identifier

ALL_MODELS = [
  Model('QL-500',     (295, 11811), compression=False, mode_setting=False, expanded_mode=False, cutting=False),
  Model('QL-550',     (295, 11811), compression=False, mode_setting=False),
  Model('QL-560',     (295, 11811), compression=False, mode_setting=False),
  Model('QL-570',     (150, 11811), compression=False, mode_setting=False),
  Model('QL-580N',    (150, 11811)),
  Model('QL-600',     (150, 11811)),
  Model('QL-650TD',   (295, 11811)),
  Model('QL-700',     (150, 11811), compression=False, mode_setting=False),
  Model('QL-710W',    (150, 11811)),
  Model('QL-720NW',   (150, 11811)),
  Model('QL-800',     (150, 11811), two_color=True, compression=False, num_invalidate_bytes=400),
  Model('QL-810W',    (150, 11811), two_color=True, num_invalidate_bytes=400),
  Model('QL-820NWB',  (150, 11811), two_color=True, num_invalidate_bytes=400),
  Model('QL-1050',    (295, 35433), number_bytes_per_row=162, additional_offset_r=44),
  Model('QL-1060N',   (295, 35433), number_bytes_per_row=162, additional_offset_r=44),
  Model('QL-1100',    (301, 35434), number_bytes_per_row=162, additional_offset_r=44),
  Model('QL-1100NWB', (301, 35434), number_bytes_per_row=162, additional_offset_r=44),
  Model('QL-1115NWB', (301, 35434), number_bytes_per_row=162, additional_offset_r=44),
]

class ModelsManager(ElementsManager):
    """
    Class for accessing the list of supported printer models
    """
    elements = copy.copy(ALL_MODELS) #: :meta private:
    element_name = 'model'
